﻿using Infragistics;
using Infragistics.Controls.Charts;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Threading;

namespace ExtremeDataChart
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();

            SyncLink.FastItemsSourceFactory = (ds) =>
            {
                _mass = new MassiveFastItemsSource((string)ds, 13, 100000000);
                return _mass;
            };

            CategoryXAxis xAxis = new CategoryXAxis();
            xAxis.ItemsSource = "d:/tmp/largeData/largeData-{0}.txt";
            xAxis.Label = "{Date}";

            NumericYAxis yAxis = new NumericYAxis();

            RangeAreaSeries area = new RangeAreaSeries();
            area.XAxis = xAxis;
            area.YAxis = yAxis;
            area.ItemsSource = "d:/tmp/largeData/largeData-{0}.txt";
            area.LowMemberPath = "Low";
            area.HighMemberPath = "High";

            theChart.HorizontalZoomable = true;
            theChart.VerticalZoomable = true;
            theChart.Axes.Add(xAxis);
            theChart.Axes.Add(yAxis);
            theChart.Series.Add(area);

            theChart.ActualWindowRectChanged += theChart_ActualWindowRectChanged;
            theChart.GridAreaRectChanged += theChart_GridAreaRectChanged;
            theChart.WindowRectMinWidth = 0.00000000001;
            theChart.WindowResponse = WindowResponse.Immediate;
        }

        void theChart_GridAreaRectChanged(object sender, RectChangedEventArgs e)
        {
            _mass.ViewportRect = e.NewRect;
        }

        void theChart_ActualWindowRectChanged(object sender, RectChangedEventArgs e)
        {
            _mass.WindowRect = e.NewRect;
        }

        private MassiveFastItemsSource _mass;
    }

    public class BucketItem
    {
        public DateTime Date { get; set; }
        public double Min { get; set; }
        public double Max { get; set; }
    }

    public class MassiveLayer
    {
        public MassiveLayer()
        {
            Buckets = new Dictionary<int, List<BucketItem>>();
        }

        public string Path { get; set; }
        public long Count { get; set; }
        public long BucketSize { get; set; }
        public int LayerIndex { get; set; }

        public BucketItem GetItem(int index)
        {
            int bucket = (int)Math.Floor((double)index / BucketSize);
            EnsureBucket(bucket);
            return Buckets[bucket][index % (int)BucketSize];
        }

        private FileStream _stream = null;
        private BinaryReader _reader = null;

        private void EnsureBucket(int bucket)
        {
            if (_stream == null)
            {
                _stream = new FileStream(Path, FileMode.Open);
                _reader = new BinaryReader(_stream);
            }

            if (Buckets.ContainsKey(bucket))
            {
                return;
            }

            int rowWidth = 24;
            if (LayerIndex == 0)
            {
                rowWidth = 16;
            }

            _stream.Seek(bucket * BucketSize * rowWidth, SeekOrigin.Begin);
            
            var newBucket = new List<BucketItem>();
            for (var i = 0; i < BucketSize; i++)
            {
                if (_stream.Position >= _stream.Length)
                {
                    break;
                }
                var newItem = new BucketItem();
                newItem.Date = DateTime.FromBinary(_reader.ReadInt64());
                newItem.Min = _reader.ReadDouble();
                if (LayerIndex == 0)
                {
                    newItem.Max = newItem.Min;
                }
                else
                {
                    newItem.Max = _reader.ReadDouble();
                }

                newBucket.Add(newItem);
            }

            Buckets[bucket] = newBucket;
        }

        public Dictionary<int, List<BucketItem>> Buckets { get; set; }
    }

    public class MassiveFastItemsSource
        : IFastItemsSource
    {
        public List<MassiveLayer> Layers { get; set; }
        public int NumLevels { get; set; }

        private MassiveLayer _layer;
        private int _currentLayer;
        public int CurrentLayer 
        {
            get
            {
                return _currentLayer;
            }
            set
            {
                var oldValue = _currentLayer;
                _currentLayer = value;
                _layer = Layers[_currentLayer];
                _minColumn.CurrentLayer = _layer;
                _maxColumn.CurrentLayer = _layer;
                System.Diagnostics.Debug.WriteLine("current layer: " + _currentLayer);
                System.Diagnostics.Debug.WriteLine("count: " + _layer.Count);
                if (oldValue != _currentLayer)
                {
                    if (Event != null)
                    {
                        Event(this, new FastItemsSourceEventArgs(FastItemsSourceEventAction.Reset, 0 , 0));
                    }
                }
            }
        }

        public MassiveFastItemsSource(string path, int numLevels, long maxCount)
        {
            _minColumn = new MassiveDoubleColumn(this, false);
            _maxColumn = new MassiveDoubleColumn(this, true);
            Layers = new List<MassiveLayer>();
            NumLevels = numLevels;

            var total = maxCount;
            for (var i = 0; i < numLevels; i++)
            {
                var layer = new MassiveLayer() { Path = string.Format(path, i), Count = maxCount };
                layer.BucketSize = 1024;
                layer.LayerIndex = i;
                Layers.Add(layer);
                maxCount = (long)Math.Round((double)maxCount / 2.0);
            }
            CurrentLayer = numLevels - 1;
            
            double max = double.MinValue;
            double min = double.MaxValue;

            for (var i = 0; i < _layer.Count; i++)
            {
                max = Math.Max(max, _layer.GetItem(i).Max);
                min = Math.Min(min, _layer.GetItem(i).Min);
            }

            _minColumn.SetMax(max);
            _minColumn.SetMin(min);

            _maxColumn.SetMax(max);
            _maxColumn.SetMin(min);
        }

        private Rect _windowRect = Rect.Empty;
        public Rect WindowRect
        {
            get
            {
                return _windowRect;
            }
            set
            {
                _windowRect = value;
                UpdateLevel();
            }
        }

        public Rect _viewportRect = Rect.Empty;
        public Rect ViewportRect
        {
            get
            {
                return _viewportRect;
            }
            set
            {
                _viewportRect = value;
                UpdateLevel();
            }
        }

        private void UpdateLevel()
        {
            if (_viewportRect.IsEmpty)
            {
                return;
            }
            if (_windowRect.IsEmpty)
            {
                return;
            }

            var extent = _viewportRect.Width / _windowRect.Width;
            for (var i = 0; i < NumLevels; i++)
            {
                if (Layers[i].Count < extent)
                {
                    CurrentLayer = i;
                    
                    break;
                }
            }
        }

        public object this[int i]
        {
            get { return _layer.GetItem(i); }
        }

        public int IndexOf(object item)
        {
            return 0;
        }

        private MassiveDoubleColumn _minColumn;
        private MassiveDoubleColumn _maxColumn;

        public IFastItemColumn<double> RegisterColumn(string propertyName)
        {
            if (propertyName == "Low")
            {
                return _minColumn;
            }
            else
            {
                return _maxColumn;
            }
        }

        public void DeregisterColumn(IFastItemColumnPropertyName fastItemColumn)
        {

        }

        public IFastItemColumn<int> RegisterColumnInt(string propertyName)
        {
            throw new NotImplementedException();
        }

        public IFastItemColumn<object> RegisterColumnObject(string propertyName)
        {
            throw new NotImplementedException();
        }

        public IFastItemColumn<DateTime> RegisterColumnDateTime(string propertyName)
        {
            throw new NotImplementedException();
        }

        public event EventHandler<FastItemsSourceEventArgs> Event;

        public System.Collections.IList ContentsAsIList()
        {
            throw new NotImplementedException();
        }

        public int Count
        {
            get { return (int)_layer.Count; }
        }

        private Dispatcher _dispatcher = null;
        public System.Windows.Threading.Dispatcher Dispatcher
        {
            get
            {
                return _dispatcher;
            }
            set
            {
                _dispatcher = value;
            }
        }
    }

    public class MassiveDoubleColumn
        : IFastItemColumn<double>
    {
       
        private MassiveFastItemsSource _source;
        public MassiveDoubleColumn(MassiveFastItemsSource source, bool isMax)
        {
            _isMax = isMax; ;
            _source = source;
        }

        private MassiveLayer _currentLayer;
        public MassiveLayer CurrentLayer
        {
            get
            {
                return _currentLayer;
            }
            set
            {
                _currentLayer = value;
            }
        }

        private double _min;
        public void SetMin(double min)
        {
            _min = min;
        }
        private double _max;
        public void SetMax(double max)
        {
            _max = max;
        }

        public double Minimum
        {
            get { return _min; }
        }

        public double Maximum
        {
            get { return _max; }
        }

        public int IndexOf(double item)
        {
            return 0;
        }

        public void Insert(int index, double item)
        {
            throw new NotImplementedException();
        }

        public void RemoveAt(int index)
        {
            throw new NotImplementedException();
        }

        private bool _isMax;

        public double this[int index]
        {
            get
            {
                if (_isMax)
                {
                    return _currentLayer.GetItem(index).Max;
                }
                else
                {
                    return _currentLayer.GetItem(index).Min;
                }
            }
            set
            {
                throw new NotImplementedException();
            }
        }

        public void Add(double item)
        {
            throw new NotImplementedException();
        }

        public void Clear()
        {
            throw new NotImplementedException();
        }

        public bool Contains(double item)
        {
            throw new NotImplementedException();
        }

        public void CopyTo(double[] array, int arrayIndex)
        {
            throw new NotImplementedException();
        }

        public int Count
        {
            get { return (int)_currentLayer.Count; }
        }

        public bool IsReadOnly
        {
            get { return true; }
        }

        public bool Remove(double item)
        {
            throw new NotImplementedException();
        }

        public IEnumerator<double> GetEnumerator()
        {
            throw new NotImplementedException();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }

        public string PropertyName
        {
            get
            {
                if (_isMax)
                {
                    return "High";
                }
                else
                {
                    return "Low";
                }
            }
        }
    }
}
